#include <Adafruit_NeoPixel.h>
#include <math.h>

// Display a waving flag on a 48x16 LED array consisting of three 16x16 WS2812b arrays.
// This version uses separate data pins for each panel, so that less memory is needed by the Neopixel library, and Arduino Uno can be used.
// compute new displacements based on a differential equation for damped harmonic oscillator: y'' + by' + y = F where the driving force F is random noise
// this leads to the update formula y(n+1) = y(n)*(1-dt^2) + (y(n)-y(n-1))*(1-b*dt) + F
// where F is the driving force, in this case random noise.

#define PIN      5  // change to whatever pin you are using to send data to the LED array.  This is now replaced by the array panelPin[]
#define N_LEDS 256  // number of LEDs in one panel.
#define imageRows 16  // the size of the image that is going to be displayed.  Height normally 16.
// number of rows and columns in each LED panel
#define numRows 16
#define numCols 16
// number of rows and columns in the flag image
#define flagRows 48
#define flagCols 96
#define brightness 70  // 255 would be maximum brightness.  I intend to use this at night, so I set the brightness much lower than the maximum.
#define ncycle 3  // *** this is the number of wave cycles across the display.  Set equal to the number of panels.

Adafruit_NeoPixel strip = Adafruit_NeoPixel(N_LEDS, PIN, NEO_GRB + NEO_KHZ800);

int numPanels = 3;  // *** 1 for single panel, 3 for 3 panels
int panelPin[3] = {5,6,7};  // Panel 0 connected to pin 5, panel 1 to pin 6, panel 2 to pin 7.
int displayCols = numCols*numPanels;
int pixel = 0;
int imageRow = 0;
int imageCol = 0;
int r=0;
int g=0;
int b=0;
int shade=128;
int x=0;
int y=0;
int xd = 0;
// int xdOff = displayCols/ncycle/4;  // offset between displacement and shading
int px = 0;
bool yEven = true;
int panel = 0;
int value = 0;
uint32_t c = 0;  // used for storing color values
// long randValue;  // for later use, not used in this sketch
float ymag = 2.0;  // magnitude of wave in vertical direction (pixels)
float slope = 5.0;  // slope of wave - sinewave will be offset horizontally by 1 pixel every 5th row.
float fcycle = float(ncycle)*2.0*3.141592654/float(displayCols);  // this is the size of a pixel increment, in radians of the wave cycle 
int yoff = int(ymag*8.0);  // integer value for later calculations
int shadeScale = int(2.5*6.0/ymag);  // controls the depth of the shading effect.

byte ysine[16][48];  // sinewave values
long newsine[16][3];  // store the two most recent values, plus the newly computed ones, in higher resolution.  Don't have enough memory to save all 16x48 as long.
long noiseMax = long(ymag*256.0*512.0*16.0*0.3);
long damping = long(512.0*0.981);  // this is the factor (1-b*dt)*512.  b is the damping coefficient.
// long noiseMax = 1L;  // these lines were used to generate continuous smooth sinewave with no noise, for debugging purposes
// long damping = long(512.0);  // this is the factor (1-b*dt)*512.  b is the damping coefficient.
long step = long((1.0-fcycle*fcycle)*512.0);  // this is the factor (1-dt^2)*512

const uint8_t flag[flagRows][flagCols][3] PROGMEM = 
#include "Canada_flag_96x48.h"   // Canada flag image.  The image is twice the resolution of the LED display, plus 8 pixels top and bottom to allow margin for waving.
;

void setup() {
  strip.begin();
  for (panel = 0; panel < numPanels; panel++) {
    strip.setPin(panelPin[panel]);
    strip.show();         // Turn OFF all pixels ASAP
  }
  strip.setBrightness(brightness); // Set BRIGHTNESS

//  Serial.begin(9600);

// precompute the sinewave pattern of pixel offsets
  for (x = 0; x<displayCols; x++) {
    for (y=0; y<numRows; y++) {
      ysine[y][x] = byte((sin((float(x)-float(y)/slope)*fcycle))*ymag*16.0+128.5);
    }
  }
  // initial values for flag displacements
  for (x = 0; x<3; x++) {
    for (y=0; y<numRows; y++) {
      newsine[y][x] = long((sin((float(x)-float(y)/slope)*fcycle))*ymag*16.0*512.0+0.5);
    }
  }
// start serial monitor, for debugging if necessary
//  Serial.print("newsine[0][0] = ");
//  Serial.println(newsine[0][0]);

}

void loop() {
  
  // waving flag animation

  // update the display, one pixel at a time
  
   for(panel=0; panel<numPanels; panel++) {
    yEven = true;
    px = panel*numCols;
    for (y=0; y<numRows; y++) {
      for (x=0; x<numCols; x++) {
        xd = x + px;  // x coordinate of the LED in the full 48x16 array
//        xd2 = xd + xdOff;  // the shading is shifted from the displacement by 1/4 of a cycle.  Maximum shading occurs at maximum slope. 
//        if (xd2 >= displayCols) {
//          xd2 -= displayCols;
//        }
        value = int((ysine[y][xd])>>4) - 8;
        if (xd<displayCols) {
          shade = 64 + (int(ysine[y][xd])-int(ysine[y][xd+1]))*shadeScale;  // Shading is used to make the waving look more realistic.  shade=128 is full brightness.
        }
        else {
          shade = 64 + (int(ysine[y][xd-1])-int(ysine[y][xd]))*shadeScale;
        }
        if (shade>128) shade = 128;
        if (shade<32) shade = 32;
        imageRow = flagRows - y*2 + value - 8;  // subtract 8 because the flag image is 16 pixels larger than twice the display size
        imageCol = xd*2 + (value>>1) + 0;  // *** adding 32 because I am using only one panel and I want the maple leaf centered.  Remove this for 3 panels.
        if ((imageCol < 0)) {  // Draw red (past the left edge of the flag image).  Change to black or a different color if your flag is not red on the left edge.
//          r = 0;
//          g = 0;
//          b = 100;
          r = (213*shade)>>7;  // multiply by shade value (which is maximum 128) and divide by 128.
          g = (43*shade)>>7;
          b = (30*shade)>>7;
        }
        else if (imageCol >= flagCols) {  // draw red (past the right edge of the flag image.  Change to black or a different color if your flag is not red on the right edge.
          r = (213*shade)>>7;
          g = (43*shade)>>7;
          b = (30*shade)>>7;
        }
        else {  // draw flag
          r = (int(pgm_read_byte_near(&flag[imageRow][imageCol][0]))*shade)>>7;
          g = (int(pgm_read_byte_near(&flag[imageRow][imageCol][1]))*shade)>>7;
          b = (int(pgm_read_byte_near(&flag[imageRow][imageCol][2]))*shade)>>7;
        }
        if (yEven) {  // y is even.  Pixel numbers increment in forward order.  Change this section if your panel is oriented differently from mine.
                      // You might need to interchange x and y, or use (numCols-x) instead of x, or (numRows-y) instead of y, to get the image in the correct orientation.
                      // If you need to interchange x and y you will also have to change "yEven" to represent whether x is even or odd.  Do this by moving the statement "yEven = !yEven;" up one line so that it is inside the x loop.
          pixel = y*numCols + x;
        }
        else {  // y is odd.  Pixel numbers decrease with x (reverse part of serpentine path)
          pixel = (y+1)*numCols - x - 1;
        }
        c = strip.Color(r, g, b);
        strip.setPixelColor(pixel , c); // Draw new pixel
      }
      yEven = !yEven;  // If your panel is rotated 90 degreed from mine, which is really the correct orientation, move this statement up one line so that it is inside the x loop.
    }
    strip.setPin(panelPin[panel]);
    strip.show();
  }
    delay(0);  // *** you can change this number to change the scrolling speed.  Smaller number will be faster scrolling.  Use 20 for single panel, 0 for 3 panels.

  // update displacement values
  // First shift the old values and compute the random noise
  for (y=0; y<numRows; y++) {
    newsine[y][2] = newsine[y][1];
    newsine[y][1] = newsine[y][0];
    newsine[y][0] = random(-noiseMax,noiseMax);
  }
  // smooth out the noise values
  for (y=0; y<numRows-1; y++) {
    newsine[y][0] = (newsine[y][0]+newsine[y+1][0])>>1;
  }
  for (y=numRows-1; y>0; y--) {
    newsine[y][0] = (newsine[y][0]+newsine[y-1][0])>>1;
  }
  for (y=0; y<numRows-1; y++) {
    newsine[y][0] = (newsine[y][0]+newsine[y+1][0])>>1;
  }
  for (y=numRows-1; y>0; y--) {
    newsine[y][0] = (newsine[y][0]+newsine[y-1][0])>>1;
  }
  // now compute the next iteration and add to the noise values
  for (y=0; y<numRows; y++) {
    newsine[y][0] = (newsine[y][0] + newsine[y][1]*step + (newsine[y][1]-newsine[y][2])*damping)>>9;
  }
  // finally, average the new iteration
  for (y=0; y<numRows-1; y++) {
    newsine[y][0] = (newsine[y][0]+newsine[y+1][0])>>1;
  }
  for (y=numRows-1; y>0; y--) {
    newsine[y][0] = (newsine[y][0]+newsine[y-1][0])>>1;
  }
  for (y=0; y<numRows-1; y++) {
    newsine[y][0] = (newsine[y][0]+newsine[y+1][0])>>1;
  }
  for (y=numRows-1; y>0; y--) {
    newsine[y][0] = (newsine[y][0]+newsine[y-1][0])>>1;
  }
  // now shift the values in ysine and transfer the new values into the first column
  for (y=0; y<numRows; y++) {
    for (x=displayCols-1; x>0; x--) {
       ysine[y][x]=ysine[y][x-1];
    }
//    randValue = newsine[y][0];
    ysine[y][0] = byte(((newsine[y][0])>>9) + 128L);
  }
// prints used for debugging
//  Serial.print("newsine[0][0] = ");
//  Serial.print(newsine[0][0]);
//  Serial.print("   ysine[0][0] = ");
//  Serial.println(ysine[0][0]);

}

